import re
import fnmatch
import os
from pprint import PrettyPrinter

# member patterns
func_pattern      = re.compile(r'\s*(final\s+)?(?P<scope>private\s+|public\s+|internal\s+)?(final\s+)?(?P<static>class\s|static\s+|mutating\s+)?(?P<type>func)\s+(?P<name>[a-zA-Z_][a-zA-Z0-9_]*\b)(?P<rest>[^{]*)')
init_pattern      = re.compile(r'\s*(final\s+)?(?P<scope>private\s+|public\s+|internal\s+)?(final\s+|convenience\s+)*(?P<type>init\??)\s*(?P<rest>[^{]*)')
var_pattern       = re.compile(r'\s*(final\s+)?(?P<add_scope>private\s*\(set\)\s+|private\s*\(get\)\s+)?(?P<scope>private\s+|public\s+|internal\s+)?(final\s+)?(?P<static>static\s+)?(?P<type>var\s+|let\s+)(?P<name>[a-zA-Z_][a-zA-Z0-9_]*\b)(?P<rest>[^{]*)(?P<computed>\s*{\s*)?')
proto_var_pattern = re.compile(r'\s*(?P<static>static\s+)?(?P<type>var\s+)(?P<name>[a-zA-Z_][a-zA-Z0-9_]*\b)(?P<rest>[^{]*)(?P<computed>\s*{(?:\s*get\s+set\s*|\s*get\s*|\s*set\s*)}\s*)?')
case_pattern      = re.compile(r'\s*(?P<type>case)\s+(?P<name>[a-zA-Z_][a-zA-Z0-9_]*\b)(\s*(?P<assoc_type>\([a-zA-Z_[(][a-zA-Z0-9_<>[\]()?!:, \t-]*\))\s*)?(\s*=\s*(?P<raw_value>.*))?')

# markdown doc patterns
param_pattern   = re.compile(r'^\s*- [pP]arameter\s*(?P<param>[^:]*):\s*(?P<desc>.*)')
return_pattern  = re.compile(r'^\s*- [rR]eturn[s]?\s*:\s*(?P<desc>.*)')
throws_pattern  = re.compile(r'^\s*- [tT]hrow[s]?\s*:\s*(?P<desc>.*)')
default_pattern = re.compile(r'^\s*- [dD]efault[s]?\s*:\s*(?P<desc>.*)')


# signatures
def class_sig(name=r'[a-zA-Z_][a-zA-Z0-9_]*'):
    return re.compile(r'\s*(final\s+)?(?P<scope>private\s+|public\s+|internal\s+)?(final\s+)?(?P<struct>class)\s+(?!func)(?P<name>' + name + r'\b)(\s*:\s*(?P<type>[^{]*))*')


def enum_sig(name=r'[a-zA-Z_][a-zA-Z0-9_]*'):
    return re.compile(r'\s*(final\s+)?(?P<scope>private\s+|public\s+|internal\s+)?(final\s+)?(?P<struct>enum)\s+(?P<name>' + name + r'\b)(\s*:\s*(?P<type>[^{]*))*')


def struct_sig(name=r'[a-zA-Z_][a-zA-Z0-9_]*'):
    return re.compile(r'\s*(final\s+)?(?P<scope>private\s+|public\s+|internal\s+)?(final\s+)?(?P<struct>struct)\s+(?P<name>' + name + r'\b)(\s*:\s*(?P<type>[^{]*))*')


def protocol_sig(name=r'[a-zA-Z_][a-zA-Z0-9_]*'):
    return re.compile(r'\s*(?P<scope>private\s+|public\s+|internal\s+)?(?P<struct>protocol)\s+(?P<name>' + name + r'\b)(\s*:\s*(?P<type>[^{]*))*')


def extension_sig(name=r'[a-zA-Z_][a-zA-Z0-9_]*'):
    return re.compile(r'\s*(?P<scope>private\s+|public\s+|internal\s+)?(?P<struct>extension)\s+(?P<name>' + name + r'\b)(\s*:\s*(?P<type>[^{]*))*(\s*where\s+(?P<where>[^{]*))?')


# debug printer
pp = PrettyPrinter(indent=1)


def pprint(*args):
    pp.pprint(*args)

# brace balancing for determining in which depth we are
string_pattern = re.compile(r'"(?:[^"\\]*(?:\\.)?)*"')
line_comment_pattern = re.compile(r'(// .*$)')
comment_pattern = re.compile(r'/\*(?:.)*\*/')


def balance_braces(line, brace_count):
    line = string_pattern.sub("", line)
    line = comment_pattern.sub("", line)
    line = line_comment_pattern.sub("", line)
    open_braces = line.count('{')
    close_braces = line.count('}')
    braces = brace_count + open_braces - close_braces
    return braces


# fetch documentation block
def get_doc_block(content, line):
    # search upwards for documentation lines
    doc_block = []
    for i in range(line, 0, -1):
        l = content[i].strip()
        if l.startswith('///'):
            converted = l[4:].rstrip()
            converted = converted.replace('`', '``')
            doc_block.insert(0, converted)
            continue
        break
    return doc_block


def doc_block_to_rst(doc_block):
    def emit_item(item):
        if item[0] == 'param':
            return ':parameter ' + item[1] + ': ' + item[2]
        elif item[0] == 'return':
            return ':returns: ' + item[1]
        elif item[0] == 'throws':
            return ':throws: ' + item[1]
        elif item[0] == 'defaults':
            return ':defaults: ' + item[1]

    last_item = None
    for l in doc_block:
        match = param_pattern.match(l)
        if match:
            if last_item:
                yield emit_item(last_item)
            match = match.groupdict()
            last_item = ['param', match['param'], match['desc']]
            continue
        match = return_pattern.match(l)
        if match:
            if last_item:
                yield emit_item(last_item)
            match = match.groupdict()
            last_item = ['return', match['desc']]
            continue
        match = throws_pattern.match(l)
        if match:
            if last_item:
                yield emit_item(last_item)
            match = match.groupdict()
            last_item = ['throws', match['desc']]
            continue
        match = default_pattern.match(l)
        if match:
            if last_item:
                yield emit_item(last_item)
            match = match.groupdict()
            last_item = ['defaults', match['desc']]
            continue
        if last_item and l == '':
            yield emit_item(last_item)
            yield ''
            last_item = None
            continue
        if not last_item:
            yield l
            continue
        last_item[len(last_item) - 1] += ' ' + l.strip()

    if last_item:
        yield emit_item(last_item)


class SwiftFileIndex(object):
    symbol_signatures = [class_sig(), enum_sig(), struct_sig(), extension_sig(), protocol_sig()]

    def __init__(self, search_path):
        self.index = []

        # find all files
        self.files = []
        for path in search_path:
            for root, dirnames, filenames in os.walk(path):
                for filename in fnmatch.filter(filenames, '*.swift'):
                    self.files.append(os.path.join(root, filename))

        for file in self.files:
            print("Indexing swift file: %s" % file)
            symbol_stack = []
            braces = 0
            with open(file, "r") as fp:
                content = fp.readlines()
                for (index, line) in enumerate(content):
                    braces = balance_braces(line, braces)

                    # track boxed context
                    for pattern in self.symbol_signatures:
                        match = pattern.match(line)
                        if match:
                            match = match.groupdict()

                            struct = match['struct'].strip()
                            if 'scope' in match and match['scope']:
                                scope = match['scope'].strip()
                            else:
                                if struct == 'extension':
                                    scope = 'public'
                                else:
                                    scope = 'internal'
                            item = {
                                'file': file,
                                'line': index,
                                'depth': braces,
                                'type': struct,
                                'scope': scope,
                                'name': match['name'].strip(),
                                'docstring': get_doc_block(content, index - 1),
                                'param': match['type'].strip() if match['type'] else None,
                                'where': match['where'].strip() if 'where' in match and match['where'] else None,
                                'children': [],
                                'raw': line
                            }
                            if len(symbol_stack) > 0 and braces > symbol_stack[-1]['depth']:
                                symbol_stack[-1]['children'].append(item)
                            else:
                                symbol_stack.append(item)

                            # find members
                            start = index
                            if line.rstrip()[-1] == '{':
                                start = index + 1
                            else:
                                for i in range(index + 1, len(content)):
                                    l = content[i].lstrip()
                                    if len(l) > 0 and l[0] == '{':
                                        start = i
                                        break
                            item['members'] = SwiftObjectIndex(content, start, item['type'])

            self.index.extend(symbol_stack)

    def find(self, name, index=None, name_prefix=[]):
        if not index:
            index = self.index

        for item in index:
            item_name = '.'.join([*name_prefix, item['name']])
            if name == item_name:
                yield item
            if len(item['children']) > 0:
                for result in self.find(name, index=item['children'], name_prefix=[*name_prefix, item['name']]):
                    yield result

    def by_file(self, index=None):
        result = {}

        if not index:
            index = self.index

        for item in index:
            if item['file'] not in result:
                result[item['file']] = []
            result[item['file']].append(item)

        return result

    @staticmethod
    def documentation(item, indent="    ", noindex=False, nodocstring=False, location=False):
        if item['param']:
            line = '.. swift:' + item['type'] + ':: ' + item['name'] + ' : ' + item['param']
        else:
            line = '.. swift:' + item['type'] + ':: ' + item['name']

        if item['where']:
            line += ' where ' + item['where']

        yield line

        if noindex:
            yield indent + ':noindex:'
        yield ''

        if not nodocstring:
            for line in doc_block_to_rst(item['docstring']):
                yield indent + line
            yield ''

        if location:
            yield indent + 'Defined in :doc:`' + item['file'] + '`:' + str(item['line'])
            yield ''


class SwiftObjectIndex(object):

    def __init__(self, content, line, typ):
        signatures = [func_pattern, init_pattern, var_pattern]
        if typ == 'enum':
            signatures = [func_pattern, init_pattern, case_pattern]
        elif typ == 'protocol':
            signatures = [func_pattern, init_pattern, proto_var_pattern]

        self.index = []
        braces = 1
        for i in range(line, len(content)):
            l = content[i]

            # balance braces
            old_braces = braces
            braces = balance_braces(l, braces)
            if braces <= 0:
                break
            if braces > 1 and old_braces == braces:
                continue

            for pattern in signatures:
                match = pattern.match(l)
                if match:
                    match = match.groupdict()
                    if 'scope' in match:
                        if match['scope']:
                            scope = match['scope'].strip()
                        else:
                            if typ == 'protocol':
                                scope = 'public'
                            else:
                                scope = 'internal'
                    else:
                        scope = 'public'
                    self.index.append({
                        'scope': scope,
                        'line': i,
                        'type': match['type'].strip(),
                        'name': match['name'].strip() if match['type'] != 'init' and match['type'] != 'init?' else 'init',
                        'static': match['static'].strip() if 'static' in match and match['static'] else None,
                        'docstring': get_doc_block(content, i - 1),
                        'rest': match['rest'].strip() if 'rest' in match and match['rest'] else None,
                        'assoc_type': match['assoc_type'].strip() if 'assoc_type' in match and match['assoc_type'] else None,
                        'raw_value': match['raw_value'].strip() if 'raw_value' in match and match['raw_value'] else None,
                        'raw': l
                    })

    @staticmethod
    def documentation(item, indent="    ", noindex=False, nodocstring=False, location=None):
        sig = item['name']
        if item['rest']:
            sig += item['rest']

        if item['type'] == 'case':
            # enum case
            if item['assoc_type']:
                yield '.. swift:enum_case:: ' + sig + item['assoc_type']
            elif item['raw_value']:
                yield '.. swift:enum_case:: ' + sig + ' = ' + item['raw_value']
            else:
                yield '.. swift:enum_case:: ' + sig
        elif item['type'] == 'var' or item['type'] == 'let':
            # variables
            if item['static'] == 'static':
                yield '.. swift:static_' + item['type'] + ':: ' + sig
            else:
                yield '.. swift:' + item['type'] + ':: ' + sig
        else:
            if item['name'] == 'init' or item['name'] == 'init?':
                yield '.. swift:init:: ' + sig
            else:
                if item['static'] == 'class':
                    yield '.. swift:class_method:: ' + sig
                else:
                    yield '.. swift:method:: ' + sig

        if noindex:
            yield indent + ':noindex:'
        yield ''

        if not nodocstring:
            for line in doc_block_to_rst(item['docstring']):
                yield indent + ' ' + line
            yield ''

        if location:
            yield indent + 'Defined in :doc:`' + location + '`:' + str(item['line'])
            yield ''



# -*- coding: utf-8 -*-
"""
    sphinx.domains.swift
    ~~~~~~~~~~~~~~~~~~~

    The Swift domain.

    :copyright: Copyright 2016 by Johannes Schriewer
    :license: BSD, see LICENSE for details.
"""

import re

from docutils import nodes
from docutils.parsers.rst import directives

from sphinx import addnodes
from sphinx.roles import XRefRole
from sphinx.locale import l_
from sphinx.domains import Domain, ObjType, Index
from sphinx.directives import ObjectDescription
from sphinx.util.nodes import make_refnode
from sphinx.util.docfields import Field, GroupedField, TypedField


def _iteritems(d):
    for k in d:
        yield k, d[k]


class SwiftObjectDescription(ObjectDescription):
    option_spec = {
        'noindex': directives.flag,
    }

    def add_target_and_index(self, name_cls_add, sig, signode):
        fullname, signature, add_to_index = name_cls_add
        if 'noindex' in self.options or not add_to_index:
            return

        for char in '<>()[]:, ?!':
            signature = signature.replace(char, "-")

        # note target
        if fullname not in self.state.document.ids:
            signode['ids'].append(signature)
            self.state.document.note_explicit_target(signode)
            self.env.domaindata['swift']['objects'][fullname] = (self.env.docname, self.objtype, signature)
        else:
            objects = self.env.domaindata['swift']['objects']
            self.env.warn(
                self.env.docname,
                'duplicate object description of %s, ' % fullname +
                'other instance in ' +
                self.env.doc2path(objects[fullname][0]),
                self.lineno)


class SwiftClass(SwiftObjectDescription):

    def handle_signature(self, sig, signode):
        container_class_name = self.env.temp_data.get('swift:class')

        # split on : -> first part is class name, second part is superclass list
        parts = [x.strip() for x in sig.split(':', maxsplit=1)]

        # if the class name contains a < then there is a generic type attachment
        if '<' in parts[0]:
            class_name, generic_type = parts[0].split('<')
            generic_type = generic_type[:-1]
        else:
            class_name = parts[0]
            generic_type = None

        # did we catch a 'where' ?
        type_constraint = None
        class_parts = None
        if ' ' in class_name:
            class_parts = class_name.split(' ')
        elif '\t' in class_name:
            class_parts = class_name.split('\t')
        if class_parts:
            # if a part starts with `where` then we have a type constraint
            for index, p in enumerate(class_parts):
                if p == 'where':
                    type_constraint = " ".join(class_parts[index:]) + ": " + parts.pop()
                    class_name = " ".join(class_parts[:index])
                    break

        if class_name.count('.'):
            class_name = class_name.split('.')[-1]

        # if we have more than one part this class has super classes / protocols
        super_classes = None
        if len(parts) > 1:
            super_classes = [x.strip() for x in parts[1].split(',')]

            # if a part starts with `where` then we have a type constraint
            for index, sup in enumerate(super_classes):
                if sup == 'where':
                    type_constraint = " ".join(super_classes[index:])
                    super_classes = super_classes[:index]
                    break

        # Add class name
        signode += addnodes.desc_addname(self.objtype, self.objtype)
        signode += addnodes.desc_name(class_name, class_name)

        # if we had super classes add annotation
        if super_classes:
            children = []
            for c in super_classes:
                prefix = ', ' if c != super_classes[0] else ''
                # ref = addnodes.pending_xref(c, reftype='protocol', refdomain='swift', reftarget=c, refwarn=True)
                ref = nodes.Text(prefix + c)
                children.append(ref)
            signode += addnodes.desc_type('', ' : ', *children)

        # add type constraint
        if type_constraint:
            signode += addnodes.desc_type(type_constraint, ' ' + type_constraint)

        add_to_index = True
        if self.objtype == 'extension' and not super_classes:
            add_to_index = False

        if container_class_name:
            class_name = container_class_name + '.' + class_name
        return self.objtype + ' ' + class_name, class_name, add_to_index

    def before_content(self):
        if self.names:
            self.env.temp_data['swift:class'] = self.names[0][1]
            self.env.temp_data['swift:class_type'] = self.objtype
            self.clsname_set = True

    def after_content(self):
        if self.clsname_set:
            self.env.temp_data['swift:class'] = None
            self.env.temp_data['swift:class_type'] = None


class SwiftClassmember(SwiftObjectDescription):

    doc_field_types = [
        TypedField('parameter', label=l_('Parameters'),
                   names=('param', 'parameter', 'arg', 'argument'),
                   typerolename='obj', typenames=('paramtype', 'type')),
        GroupedField('errors', label=l_('Throws'), rolename='obj',
                     names=('raises', 'raise', 'exception', 'except', 'throw', 'throws'),
                     can_collapse=True),
        Field('returnvalue', label=l_('Returns'), has_arg=False,
              names=('returns', 'return')),
    ]

    def _parse_parameter_list(self, parameter_list):
        parameters = []
        parens = {'[]': 0, '()': 0, '<>': 0}
        last_split = 0
        for i, c in enumerate(parameter_list):
            for key, value in parens.items():
                if c == key[0]:
                    value += 1
                    parens[key] = value
                if c == key[1]:
                    value -= 1
                    parens[key] = value

            skip_comma = False
            for key, value in parens.items():
                if value != 0:
                    skip_comma = True

            if c == ',' and not skip_comma:
                parameters.append(parameter_list[last_split:i].strip())
                last_split = i + 1
        parameters.append(parameter_list[last_split:].strip())

        result = []
        for parameter in parameters:
            name, rest = [x.strip() for x in parameter.split(':', maxsplit=1)]
            name_parts = name.split(' ', maxsplit=1)
            if len(name_parts) > 1:
                name = name_parts[0]
                variable_name = name_parts[1]
            else:
                name = name_parts[0]
                variable_name = name_parts[0]
            equals = rest.rfind('=')
            if equals >= 0:
                default_value = rest[equals + 1:].strip()
                param_type = rest[:equals].strip()
            else:
                default_value = None
                param_type = rest
            result.append({
                "name": name,
                "variable_name": variable_name,
                "type": param_type,
                "default": default_value
            })
        return result

    def handle_signature(self, sig, signode):
        container_class_name = self.env.temp_data.get('swift:class')
        container_class_type = self.env.temp_data.get('swift:class_type')

        # split into method name and rest
        first_anglebracket = sig.find('<')
        first_paren = sig.find('(')
        if first_anglebracket >= 0 and first_paren > first_anglebracket:
            split_point = sig.find('>')
        else:
            split_point = first_paren
        method_name = sig[0:split_point]

        # find method specialization
        angle_bracket = method_name.find('<')
        if angle_bracket >= 0:
            method_name = method_name[:angle_bracket]

        rest = sig[split_point:]

        # split parameter list
        parameter_list = None
        depth = 0
        for i, c in enumerate(rest):
            if c == '(':
                depth += 1
            elif c == ')':
                depth -= 1
            if depth == 0:
                parameter_list = rest[1:i]
                rest = rest[i + 1:]
                break

        if len(parameter_list) > 0:
            parameters = self._parse_parameter_list(parameter_list)
        else:
            parameters = []

        # check if it throws
        throws = rest.find('throws') >= 0

        # check for return type
        return_type = None
        arrow = rest.find('->')
        if arrow >= 0:
            return_type = rest[arrow + 2:].strip()

        # build signature and add nodes
        signature = ''
        if self.objtype == 'static_method':
            signode += addnodes.desc_addname("static", "static func")
        elif self.objtype == 'class_method':
            signode += addnodes.desc_addname("class", "class func")
        elif self.objtype != 'init':
            signode += addnodes.desc_addname("func", "func")

        if self.objtype == 'init':
            signode += addnodes.desc_name('init', 'init')
            signature += 'init('
            for p in parameters:
                signature += p['name'] + ':'
            signature += ')'
        else:
            signode += addnodes.desc_name(method_name, method_name)
            signature += method_name
            signature += '('
            for p in parameters:
                signature += p['name'] + ':'
            signature += ')'

        params = []
        sig = ''
        for p in parameters:
            param = p['name'] + ': ' + p['type']
            sig += p['name'] + ':'
            if p['default']:
                param += ' = ' + p['default']
            params.append(addnodes.desc_parameter(param, param))

        signode += addnodes.desc_parameterlist(sig, "", *params)

        title = signature
        if throws:
            signode += addnodes.desc_annotation("throws", "throws")
            signature += "throws"

        if return_type:
            signode += addnodes.desc_returns(return_type, return_type)
            signature += "-" + return_type

        if container_class_type == 'protocol':
            signature += "-protocol"

        if self.objtype == 'static_method':
            signature += '-static'
        elif self.objtype == 'class_method':
            signature += '-class'

        if container_class_name:
            return (container_class_name + '.' + title), (container_class_name + '.' + signature), True
        return title, signature, True


class SwiftEnumCase(SwiftObjectDescription):

    def handle_signature(self, sig, signode):
        container_class_name = self.env.temp_data.get('swift:class')
        enum_case = None
        assoc_value = None
        raw_value = None

        # split on ( -> first part is case name
        parts = [x.strip() for x in sig.split('(', maxsplit=1)]
        enum_case = parts[0].strip()
        if len(parts) > 1:
            parts = parts[1].rsplit('=', maxsplit=1)
            assoc_value = parts[0].strip()
            if len(parts) > 1:
                raw_value = parts[1].strip()
            if assoc_value == "":
                assoc_value = None
            else:
                assoc_value = "(" + assoc_value
        else:
            parts = [x.strip() for x in sig.split('=', maxsplit=1)]
            enum_case = parts[0].strip()
            if len(parts) > 1:
                raw_value = parts[1].strip()

        # Add class name
        signode += addnodes.desc_name(enum_case, enum_case)
        if assoc_value:
            signode += addnodes.desc_type(assoc_value, assoc_value)
        if raw_value:
            signode += addnodes.desc_addname(raw_value, " = " + raw_value)

        if container_class_name:
            enum_case = container_class_name + '.' + enum_case
        return enum_case, enum_case, True


var_sig = re.compile(r'^\s*(?P<name>[a-zA-Z_][a-zA-Z0-9_]*\b)(\s*:\s*(?P<type>[a-zA-Z_[(][a-zA-Z0-9_<>[\]()?!:, \t-]*))?(\s*=\s*(?P<value>[^{]*))?')


class SwiftClassIvar(SwiftObjectDescription):

    doc_field_types = [
        Field('defaultvalue', label=l_('Default'), has_arg=False,
              names=('defaults', 'default')),
    ]

    def handle_signature(self, sig, signode):
        container_class_name = self.env.temp_data.get('swift:class')

        match = var_sig.match(sig)
        if not match:
            self.env.warn(
                self.env.docname,
                'invalid variable/constant documentation string "%s", ' % sig,
                self.lineno)
            return

        match = match.groupdict()

        if self.objtype == 'static_var':
            signode += addnodes.desc_addname("static var", "static var")
        elif self.objtype == 'static_let':
            signode += addnodes.desc_addname("static let", "static let")
        elif self.objtype == 'var':
            signode += addnodes.desc_addname("var", "var")
        elif self.objtype == 'let':
            signode += addnodes.desc_addname("let", "let")

        name = match['name'].strip()
        signature = name
        signode += addnodes.desc_name(name, name)
        if match['type']:
            typ = match['type'].strip()
            signature += '-' + typ
            signode += addnodes.desc_type(typ, " : " + typ)
        if match['value'] and len(match['value']) > 0:
            value = match['value'].strip()
            signode += addnodes.desc_addname(value, " = " + value)
        elif match['value']:
            signode += addnodes.desc_addname('{ ... }', ' = { ... }')

        signature += "-" + self.objtype

        if container_class_name:
            name = container_class_name + '.' + name
            signature = container_class_name + '.' + signature
        return name, signature, True


class SwiftXRefRole(XRefRole):

    def process_link(self, env, refnode, has_explicit_title, title, target):
        return title, target


class SwiftModuleIndex(Index):
    """
    Index subclass to provide the Swift module index.
    """

    name = 'modindex'
    localname = l_('Swift Module Index')
    shortname = l_('Index')

    @staticmethod
    def indexsorter(a):
        type_order = ['class', 'struct', 'enum', 'protocol', 'extension']

        for i, t in enumerate(type_order):
            if a[0].startswith(t):
                return '{:04d}{}'.format(i, a[0])
        return a[0]

    def generate(self, docnames=None):
        content = []
        collapse = 0

        entries = []
        for refname, (docname, typ, signature) in _iteritems(self.domain.data['objects']):
            info = typ.replace("_", " ")
            entries.append((
                refname,
                0,
                docname,
                signature,
                info,
                '',
                ''
            ))

        entries = sorted(entries, key=lambda x: x[3][0])
        current_list = []
        current_key = None
        for entry in entries:
            if entry[3][0].upper() != current_key:
                if len(current_list) > 0:
                    content.append((current_key, current_list))
                current_key = entry[3][0].upper()
                current_list = []
            current_list.append(entry)
        content.append((current_key, current_list))

        result = []
        for key, entries in content:
            e = sorted(entries, key=self.indexsorter)
            result.append((key, e))

        return result, collapse


class SwiftDomain(Domain):
    """Swift language domain."""
    name = 'swift'
    label = 'Swift'
    object_types = {
        'function':        ObjType(l_('function'),            'function',     'obj'),
        'method':          ObjType(l_('method'),              'method',       'obj'),
        'class_method':    ObjType(l_('class method'),        'class_method', 'obj'),
        'static_method':   ObjType(l_('static method'),       'static_method','obj'),
        'class':           ObjType(l_('class'),               'class',        'obj'),
        'enum':            ObjType(l_('enum'),                'enum',         'obj'),
        'enum_case':       ObjType(l_('enum case'),           'enum_case',    'obj'),
        'struct':          ObjType(l_('struct'),              'struct',       'obj'),
        'init':            ObjType(l_('initializer'),         'init',         'obj'),
        'protocol':        ObjType(l_('protocol'),            'protocol',     'obj'),
        'extension':       ObjType(l_('extension'),           'extension',    'obj'),
        'default_impl':    ObjType(l_('extension'),           'default_impl', 'obj'),
        'let':             ObjType(l_('constant'),            'let',          'obj'),
        'var':             ObjType(l_('variable'),            'var',          'obj'),
        'static_let':      ObjType(l_('static/class constant'),     'static_let',   'obj'),
        'static_var':      ObjType(l_('static/class variable'),     'static_var',   'obj'),
    }

    directives = {
        'function':        SwiftClassmember,
        'method':          SwiftClassmember,
        'class_method':    SwiftClassmember,
        'static_method':   SwiftClassmember,
        'class':           SwiftClass,
        'enum':            SwiftClass,
        'enum_case':       SwiftEnumCase,
        'struct':          SwiftClass,
        'init':            SwiftClassmember,
        'protocol':        SwiftClass,
        'extension':       SwiftClass,
        'default_impl':    SwiftClass,
        'let':             SwiftClassIvar,
        'var':             SwiftClassIvar,
        'static_let':      SwiftClassIvar,
        'static_var':      SwiftClassIvar,
    }

    roles = {
        'function':      SwiftXRefRole(),
        'method':        SwiftXRefRole(),
        'class':         SwiftXRefRole(),
        'enum':          SwiftXRefRole(),
        'enum_case':     SwiftXRefRole(),
        'struct':        SwiftXRefRole(),
        'init':          SwiftXRefRole(),
        'static_method': SwiftXRefRole(),
        'class_method':  SwiftXRefRole(),
        'protocol':      SwiftXRefRole(),
        'extension':     SwiftXRefRole(),
        'default_impl':  SwiftXRefRole(),
        'let':           SwiftXRefRole(),
        'var':           SwiftXRefRole(),
        'static_let':    SwiftXRefRole(),
        'static_var':    SwiftXRefRole()
    }
    initial_data = {
        'objects': {},  # fullname -> docname, objtype
    }
    indices = [
        SwiftModuleIndex,
    ]

    def clear_doc(self, docname):
        for fullname, (fn, _, _) in list(self.data['objects'].items()):
            if fn == docname:
                del self.data['objects'][fullname]

    def resolve_xref(self, env, fromdocname, builder,
                     typ, target, node, contnode):
        for refname, (docname, type, signature) in _iteritems(self.data['objects']):
            if refname == target:
                node = make_refnode(builder, fromdocname, docname, signature, contnode, target)
                return node
        return None

    def get_objects(self):
        for refname, (docname, type, signature) in _iteritems(self.data['objects']):
            yield (refname, refname, type, docname, refname, 1)


def setup(app):
    from .autodoc import SwiftAutoDocumenter, build_index
    app.add_autodocumenter(SwiftAutoDocumenter)

    app.add_domain(SwiftDomain)
    app.add_config_value('swift_search_path', ['../src'], 'env')
    app.add_config_value('autodoc_default_flags', [], True)
    build_index(app)



import argparse
import os

from swift_domain.indexer import SwiftFileIndex, SwiftObjectIndex

parser = argparse.ArgumentParser(description='Bootstrap ReStructured Text documentation for Swift code.')
parser.add_argument(
    'source_path',
    type=str,
    help='Path to Swift files'
)
parser.add_argument(
    'documentation_path',
    type=str,
    help='Path to generate the documentation in'
)
parser.add_argument(
    '--private',
    dest='private',
    action='store_true',
    help='Include private and internal members',
    required=False,
    default=False
)
parser.add_argument(
    '--overwrite',
    dest='overwrite',
    action='store_true',
    help='Overwrite existing documentation',
    required=False,
    default=False
)
parser.add_argument(
    '--undoc-members',
    dest='undoc',
    action='store_true',
    help='Include members without documentation block',
    required=False,
    default=False
)
parser.add_argument(
    '--no-members',
    dest='members',
    action='store_false',
    help='Do not include member documentation',
    required=False,
    default=True
)
parser.add_argument(
    '--file-location',
    dest='location',
    action='store_true',
    help='Add a paragraph with file location where the member was defined',
    required=False,
    default=False
)
parser.add_argument(
    '--no-index',
    dest='noindex',
    action='store_true',
    help='Do not add anything to the index',
    required=False,
    default=False
)
parser.add_argument(
    '--no-index-members',
    dest='noindex_members',
    action='store_true',
    help='Do not add members to the index, just the toplevel items',
    required=False,
    default=False
)
parser.add_argument(
    '--exclude-list',
    dest='exclusion_list',
    metavar='file',
    type=str,
    required=False,
    default=None,
    help='File with exclusion list for members'
)
parser.add_argument(
    '--use-autodocumenter',
    dest='autodocumenter',
    action='store_true',
    help='''Do not dump actual documentation but rely on the auto documenter,
    may duplicate documentation in case you have defined extensions in multiple files''',
    required=False,
    default=False
)


def main():
    args = parser.parse_args()
    source_path = os.path.abspath(args.source_path)
    file_index = SwiftFileIndex([source_path])

    try:
        os.makedirs(args.documentation_path)
    except os.FileExistsError:
        pass

    # check for overwrite
    for file, members in file_index.by_file().items():
        destfile = get_dest_file(file, args.source_path, args.documentation_path)
        if os.path.exists(destfile) and not args.overwrite:
            print("""ERROR: {} already exists, to overwrite existing
                     documentation use the '--overwrite' flag""".format(file))
            exit(1)

    exclusion_list = []
    if args.exclusion_list:
        exclusion_list = open(args.exclusion_list, 'r').readlines()

    for file, members in file_index.by_file().items():
        destfile = get_dest_file(file, args.source_path, args.documentation_path)
        print("Writing documentation for '{}'...".format(os.path.relpath(file, source_path)))
        try:
            os.makedirs(os.path.dirname(destfile))
        except os.FileExistsError:
            pass
        with open(destfile, "w") as fp:
            heading = 'Documentation for {}'.format(os.path.relpath(file, source_path))
            fp.write(('=' * len(heading)) + '\n')
            fp.write(heading + '\n')
            fp.write(('=' * len(heading)) + '\n\n\n')
            if args.autodocumenter:
                auto_document(members, args, exclusion_list, fp)
            else:
                document(members, args, exclusion_list, file, fp, '')


def get_dest_file(filename, search_path, doc_path):
    rel = os.path.relpath(filename, search_path)
    return os.path.join(doc_path, rel)[:-6] + '.rst'


def auto_document(members, args, exclusion_list, fp):
    for member in members:
        add = True
        if member['name'] in exclusion_list:
            add = False
        if args.undoc is False and len(member['docstring']) == 0:
            add = False
        if args.private is False and member['scope'] != 'public':
            add = False
        if not add:
            continue

        fp.write('.. autoswift:: {}\n'.format(member['name']))
        if args.noindex:
            fp.write('   :noindex:\n')
        if args.noindex_members:
            fp.write('   :noindex-members:\n')
        if args.members:
            fp.write('   :members:\n')
        if args.undoc:
            fp.write('   :undoc-members:\n')
        if args.location:
            fp.write('   :file-location:\n')
        if args.private:
            fp.write('   :private-members:\n')


def document(members, args, exclusion_list, file, fp, indent):
    for member in members:
        add = True
        if member['name'] in exclusion_list:
            add = False
        if args.undoc is False and len(member['docstring']) == 0:
            add = False
        if args.private is False and member['scope'] != 'public':
            add = False
        if not add:
            continue

        doc = SwiftFileIndex.documentation(
            member,
            indent=indent,
            location=args.location,
            nodocstring=args.undoc,
            noindex=args.noindex
        )
        for line in doc:
            content = indent + line + "\n"
            fp.write(content)

        if args.members:
            document_member(member, args, exclusion_list, file, fp, indent)
        fp.write('\n')


def document_member(parent, args, exclusion_list, file, fp, indent):
    for member in parent['members'].index:
        add = True
        if member['name'] in exclusion_list:
            add = False
        if args.undoc is False and len(member['docstring']) == 0:
            add = False
        if args.private is False and member['scope'] != 'public':
            add = False
        if not add:
            continue

        doc = SwiftObjectIndex.documentation(
            member,
            indent=indent,
            location=file if args.location else None,
            nodocstring=False,
            noindex=(args.noindex or args.noindex_members)
        )
        for line in doc:
            content = indent + '   ' + line + "\n"
            fp.write(content)

    document(parent['children'], args, exclusion_list, file, fp, indent + '   ')


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""`anarchytools` lives on `Github`_.

.. _github: https://www.github.com/AnarchyTools/anarchySphinx

"""
from setuptools import setup
from anarchy_theme import __version__


setup(
    name='anarchy_sphinx',
    version=__version__,
    url='https://github.com/AnarchyTools/anarchySphinx',
    license='BSD',
    author='Johannes Schriewer',
    author_email='hallo@dunkelstern.de',
    description='AnarchyTools Theme and Swift support for Sphinx.',
    long_description=open('README.rst').read(),
    zip_safe=False,
    packages=['anarchy_theme', 'swift_domain'],
    package_data={
        'anarchy_theme': [
            'theme.conf',
            '*.html',
            'static/css/*.css'
        ]
    },
    entry_points={
        'console_scripts': [
            'anarchysphinx=swift_domain.bootstrap:main',
        ],
    },
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Topic :: Documentation',
        'Topic :: Software Development :: Documentation',
    ],
)



from sphinx.ext.autodoc import Documenter, bool_option, members_option, members_set_option
from swift_domain.indexer import SwiftFileIndex, SwiftObjectIndex

file_index = None


def build_index(app):
    global file_index
    file_index = SwiftFileIndex(app.config.swift_search_path)


class SwiftAutoDocumenter(Documenter):
    objtype = 'swift'
    option_spec = {
        'noindex': bool_option,                 # do not add to index
        'noindex-members': bool_option,         # do not index members
        'members': members_option,              # document members, optional: list
        'recursive-members': bool_option,       # recursively document members
        'undoc-members': bool_option,           # include members without docstring
        'nodocstring': bool_option,             # do not show the docstring
        'file-location': bool_option,           # add a paragraph with the file location
        'exclude-members': members_set_option,  # exclude these members
        'private-members': bool_option          # show private members
    }

    def __init__(self, *args, **kwargs):
        super(SwiftAutoDocumenter, self).__init__(*args, **kwargs)
        self.append_at_end = []

    def generate(self, **kwargs):
        global file_index

        emit_warning = True
        for index in file_index.find(self.name):
            self.document(index)
            emit_warning = False

        if emit_warning:
            self.env.warn(
                self.env.docname,
                'can not find "%s" in any Swift file!' % self.name)

    def document(self, item, indent=''):

        doc = SwiftFileIndex.documentation(
            item,
            indent=self.content_indent,
            location=('file-location' in self.options),
            nodocstring=('nodocstring' in self.options),
            noindex=('noindex' in self.options)
        )
        for line in doc:
            content = indent + line
            self.add_line(content, '<autodoc>')

        # only document members when asked to
        if 'members' not in self.options:
            return

        member_list = self.options.members if isinstance(self.options.members, list) else None
        exclude_list = self.options.exclude_members if isinstance(self.options.exclude_members, list) else []
        for member in item['members'].index:
            add = False
            if not member_list or member['name'] in member_list:
                add = True
            if member['name'] in exclude_list:
                add = False
            if 'undoc-members' in self.options and len(member['docstring']) == 0:
                add = False
            if 'private-members' not in self.options and member['scope'] != 'public':
                add = False
            if add:
                loc = item['file'] if 'file-location' in self.options else None
                doc = SwiftObjectIndex.documentation(
                    member,
                    indent=self.content_indent,
                    location=loc,
                    nodocstring=('nodocstring' in self.options),
                    noindex=('noindex' in self.options or 'noindex-members' in self.options)
                )
                for line in doc:
                    content = indent + self.content_indent + line
                    self.add_line(content, '<autodoc>')

        if 'recursive-members' in self.options:
            for child in item['children']:
                self.document(child, indent=indent + self.content_indent)
